#ifndef BUFFER_H
#define BUFFER_H
#include "Basics.h"
#include <stdexcept>

class buffer
{
	public:

/* Constructors */	buffer();
					buffer(const char* bufferTitle);

/*  Destructor  */	~buffer();

/*  Accessors   */	u32 size(); // Returns sizeUsed
					char* title(); // Returns title of buffer

					// Puts data in 'dest' if 'location' is within the buffer
					template <typename valueType>
						bool get(valueType &dest, u32 location);
					bool getBit(bool &dest, u32 location, u32 bitNum);
					bool getString(char* dest, u32 location, u32 size);

					// Use the data pointer with care, don't read past dest[sizeUsed-location-1]
					const void* getPtr(u32 location);
					const void* getPtr(u32 location, u32 sizeRequested);
					bool getPtr(char* &dest, u32 location);
					template <typename valueType>
						bool getPtr(valueType* &dest, u32 location, u32 sizeRequested);

					// Returns data if the location is < sizeUsed, otherwise returns 0
					template <typename valueType>
						valueType get(u32 location);
					bool getBit(u32 location, u32 bitNum);

/*   Searching   */	// Returns true if the sequence was found at the location
					bool has(char character, u32 location);
					bool has(const char* str, u32 location, u32 size);

					// Puts next instance of 'character' in 'dest' if found
					bool getNext(char character, u32 start, u32 &dest);
					bool getNextUnquoted(char character, u32 start, u32 &dest);

/*   Mutators   */	// Adds data onto the end of the buffer
					template <typename valueType>
						bool add(valueType value);
					template <typename valueType>
						bool add(valueType value, u32 amount);
					bool addStr(const char* chunk, u32 chunkSize);

					// Inserts data at a position in the buffer
					template <typename valueType>
						bool insert(u32 location, valueType value);
					bool insertStr(u32 startLocation, const char* chunk, u32 chunkSize);

					// Replaces the specified portion of the buffer
					template <typename valueType>
						bool replace(u32 location, valueType value);
					bool replaceStr(u32 startLocation, const char* chunk, u32 chunkSize);
					bool replaceStr(u32 startLocation, u32 initialSize, const char* chunk, u32 chunkSize);

					// Swap two parts of a buffer
					template <typename valueType>
						bool swap(u32 location1, u32 location2);
					bool swapStr(u32 location1, u32 location2, u32 swapSize);

					// Replace the whole buffer with the new string
					bool overwrite(const char* chunk, u32 chunkSize);

					// Transfer all contents of source to this buffer, this buffer is flushed before transfer, source is flushed after
					bool takeAllData(buffer& source);

					// Gives the section a new name
					bool setTitle(char* newTitle);
					bool setTitle(u32 newTitle);

					// Deletes a part of the buffer
					template <typename valueType>
						bool del(u32 location);
					bool delRange(u32 startLocation, u32 endLocation);
					bool del(u32 startLocation, u32 size);
					/*bool del(u32 startLocation, u32 endLocation);
					bool delAmount();*/

/*  IO Methods  */	// Writes the data to 'pFile'
					void write(FILE* pFile);

					// Sets the title to the first four characters and extracts data the size of the following long
					bool extract(buffer &buf, u32 &position);

					friend bool FileToBuffer(void* &hMpq, const char* fileName, buffer &buf);
					friend bool FileToBuffer(const char* FileName, buffer &buf);

/*     Misc     */	// Checks whether a referance or pointer to a buffer is valid
					bool exists();
					
					// Sets the amount of data allocated for the buffer, does not change 'sizeUsed'
					bool setSize(u32 size);

					// Deletes data and sets sizeUsed and sizeAllotted to zero
					void flush();


	private:

/*	   Data	    */  char* data; // The actual data contained by the buffer
					char bufTitle[5]; // Title of the buffer
					u32 sizeAllotted; // Size allocated for data
					u32 sizeUsed; // Size of data being used

/* Priv Methods */  /** Called to increase or decrease buffer size
						If multiplySize is set, it multiplies the size
						increased to by DEFAULT_SIZE_MULTIPLIER */
					bool resize(s64 sizeChange, bool multiplySize);

					const double DEFAULT_SIZE_MULTIPLIER; // +20%

};

/**
	Thrown if allocation of extra buffer space fails,
	You can pass the number (shown below) of the
	error message you wish to have returned by a
	call to what()

	0: "could not create a temporary buffer - you may have ran out of memory.",
	1: "could not initialize buffer - you may have ran out of memory."
*/

class BadResize : public std::bad_alloc
{
	public:
		explicit BadResize(u8 inExNum) { }
		virtual const char* what() const throw();
		virtual ~BadResize() throw() { }

	private:
		int exNum;
};

#endif